/*
 * A simple parallel execution demon.
 *
 * Features:
 *     Local command interface allows local user to connect and send
 *     commands to both local and remote demons
 * 
 *     Fault-tolerance (tries to maintain itself, except for the "master")
 * 
 *     Messages between demons can use various forms of security-enhancing
 *     features, including SSL (Secure Socket Layer) (is this true) and
 *     encryption.
 *
 *     GUI state display option
 */

#include "pexec.h"
#include "msgpass.h"
#include <stdio.h>
#include <errno.h>

#define LOCAL_SOCKET "/tmp/pdemon"
#define STDIN 0

int GetConnection ANSI_ARGS((PX_Ctx *, PX_Conn *, PX_Handler, int )), 
    UnlinkLocal ANSI_ARGS((PX_Ctx *, PX_Conn *, PX_Handler, int ));
extern int ReadCommand ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int));
extern void *InitReadCommand( );

/* Message passing state.  Passed to other routines ?? */
static MSG_State  *mctx = 0;

/*
   Each demon is brought up separately.  Either from the command line 
    -master name port
   or from the local port
    master name port
   a demon can add itself to the systems known by the master.
    -port number
   can be used by the master to identify its port.

   Connections to the local socket are just like connections to stdin.
 */
int main( argc, argv )
int  argc;
char **argv;
{
PX_Ctx     ctx;
PX_Handler exit_on_close, err_fd;
PX_Handler read_command;
PX_Handler unlinklocal, getconnection;
void       *sock_info;
int        fd;
int        i;
int        portnum;

PX_Ctx_init( &ctx );
PX_Monitor_setup( &ctx );

/* Create the local socket command port */
fd = PX_FD_Create_local_socket( LOCAL_SOCKET, 1, &sock_info );
if (fd < 0) {
    fprintf( stderr, "Error creating local socket\n" );
    }
else {
    getconnection = PX_Handler_create( GetConnection, sock_info, (int (*)())0,
				   (void *)0, UnlinkLocal );
    unlinklocal   = PX_Handler_simple_create( UnlinkLocal, (void *)0 );
    PX_FD_add( fd, PX_READ, getconnection, unlinklocal, unlinklocal, 
	  (struct timeval *)0, &ctx );
    PX_Handler_free( unlinklocal );
    PX_Handler_free( getconnection );
    }

/* If command-line specifies master, start trying to connect to it */
portnum = 0;
mctx    = 0;
if (argc > 1) {
    for (i=1; i<argc; i++) {
	if (!argv[i]) continue;
	if (strcmp( argv[i], "-port" ) == 0) {
	    if (i + 1 >= argc) {
		}
	    else {
		portnum = atoi( argv[i+1] );
		i ++;
		}
	    }
	else if (strcmp( argv[i], "-master" ) == 0) {
	    if (i + 2 >= argc) {
		printf( "usage: -master host port\n" );
		}
	    else {
		printf( "Should try to connect to %s %s\n", 
		       argv[i+1], argv[i+2] );
		/* This adds a connection to ctx and returns a 
		   message context */
		MSG_Master_connect( &ctx, argv[i+1], atoi(argv[i+2]), 
				    &mctx );
		/* Skip over the name/port */
		i += 2;
		}
	    }
	}
    }
/* I might be both a master and a slave.  Leave for later */
if (!mctx && portnum > 0) {
    /* I'm a master.  Setup a network master contact.  This selects a portnum
       if one not chosen */
    MSG_Master_setup( &ctx, &portnum, &mctx );
    printf( "port for contacting master is %d\n", portnum );
    }

/* Add a command line interpreter to stdin */
/* Add a command interpreter to stdin */
read_command = PX_Handler_simple_create( ReadCommand, 
					InitReadCommand(&ctx,&mctx) );
exit_on_close = PX_Handler_simple_create( PX_Hfcn_exit_on_close, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
PX_FD_add( STDIN,  PX_READ, read_command, err_fd, exit_on_close, 
	  (struct timeval *)0, &ctx );


/* Enter PEXEC loop waiting for connections */
if (PX_Gselect( &ctx, (void *)0 )) {
    perror( "Gselect failed" );
    return 1;
    }
return 0;
}

/* This is called when a listen allows an accept */
int GetConnection( ctx, conn, handler, in_mask )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_mask;
{
int fd;
PX_Handler err_fd, close_fd, fd_to_set;

fd = PX_FD_Accept_connection( conn->fd, handler->extra_data );
if (fd < 0) {
    fprintf( stderr, "accept errno = %d\n", errno );
    perror( "Accept failed" );
    return fd;
    }

close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
fd_to_set     = PX_Handler_simple_create( PX_Hfcn_fd_to_set, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
PX_FD_add( fd, PX_READ, fd_to_set, err_fd, close_fd, (struct timeval *)0, 
	   ctx );
PX_FdAddToSet( fd_to_set, 1 );
PX_FdAddToSet( fd_to_set, 2 );
return 0;
}

int UnlinkLocal( ctx, conn, handler, err )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        err;
{
unlink( LOCAL_SOCKET );
return 0;
}
