/*
 * This is a simple command-line handler that accepts a few commands.
 * It is for example purposes only.  That is why these routines do NOT
 * use the PX_ prefix
 */
#include "pexec.h"
#include <stdio.h>
#include <memory.h>
#include <string.h>

#define MAX_BUF 128
typedef struct {
    int  curlen;
    char buf[MAX_BUF+1];
    } CommandBuf;

/*
 * This is a sample handler for accepting commands from stdin.  The commands
 * accepted are described in the code.  They use ParsePgm to create a
 * (pgm,argv) pair.
 */

int ParsePgm( );
void FreePgmArgs( );
void PrintPgmArgs( );

void *InitReadCommand( ctx )
PX_Ctx *ctx;
{
CommandBuf *b = (CommandBuf *)MALLOC( sizeof(CommandBuf) );
b->curlen = 0;

/* Turn on handling child signals */
PX_Proc_set_child_handler( ctx );

return (void *)b;
}

int ReadCommand( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
int        n, err;
CommandBuf *cbuf = (CommandBuf *)handler->extra_data;
char       *p, *p1;

/* Read a new command */
if (cbuf->curlen == MAX_BUF) {
    cbuf->buf[MAX_BUF] = 0;
    fprintf( stderr, "Discarding command %s\n", cbuf->buf );
    cbuf->curlen = 0;
    }

n = read( conn->fd, cbuf->buf+cbuf->curlen, MAX_BUF-cbuf->curlen );

if (n <= 0) 
    return PX_Hfcn_do_eof( ctx, conn, handler, n );

/* Actually read something */
cbuf->curlen += n;
cbuf->buf[cbuf->curlen] = 0;

/* Process the command.  There may be several; process them all before 
   returning */
while (1) {
    /* Is there a newline? If not, we'll just return to the select loop for the
       rest of the data */
    p = strchr( cbuf->buf, '\n' );
    if (!p) return 0;
    *p = 0;

    /* 
       Yes, there is a newline.  Process the command and update the buffer.
       The commands are:
       
       x          - exit
       r pgm args - run program
       o host pgm args - run program remotely
       d          - dump information on connections
       */
    p1 = cbuf->buf;
    /* Skip whitespace */
    while (*p1 && isspace(*p1)) p1++;
    switch(*p1) {
	case 'x': PX_Ctx_close_all_conn( ctx );
	    break;
	case 'o': 
	    {
	    char *pgm, **argv, *host;
	    int  rd_in, wrt_out, wrt_err;
	    PX_Handler to_stdout, close_fd, err_fd;
	    PX_Pgm *pgmst;
	    
	    /* get host from string */
	    p1++;
	    while (*p1 && isspace(*p1)) p1++;
	    host = p1;
	    while (*p1 && !isspace(*p1)) p1++;
	    *p1 = 0;
	    
	    /* get pgm and arguments from string */
	    if (ParsePgm( p1+1, &pgm, &argv )) {
		fprintf( stderr, "Could not parse command line %s\n", 
			p1 + 1);
		break;
		}
	    if (err = PX_Proc_run_on_host( host, pgm, argv, 
				     &rd_in, &wrt_out, &wrt_err, &pgmst )) {
		fprintf( stderr, "Could not run %s on %s\n", pgm, host );
		if (err == PX_EXEC_FAILED)
		    exit(1);
		break;
		}
	    /* Attach handlers - wrt_out/err to stdout, rd_in ignored */
	    /* Eventually, we may want to have permanent handlers for this */
	    close_fd	= PX_Handler_simple_create( PX_Hfcn_close_fd, 
						    (void *)0 );
	    to_stdout	= PX_Handler_simple_create( PX_Hfcn_to_stdout, 
						    (void *)0 );
	    err_fd	= PX_Handler_simple_create( PX_Hfcn_error_fd, 
						    (void *)0 );
	    
	    PX_FD_add( wrt_out, 1, to_stdout, err_fd, close_fd, 
		     (struct timeval *)0, ctx );
	    PX_FD_add( wrt_err, 1, to_stdout, err_fd, close_fd, 
		     (struct timeval *)0, ctx );
	    /* It would be nice to report the exit status of this program */
	    }
	    break;

	case 'r': 
	    {
	    char *pgm, **argv;
	    int  rd_in, wrt_out, wrt_err;
	    PX_Handler to_stdout, close_fd, err_fd;
	    PX_Pgm *pgmst;
	    
	    /* get pgm and arguments from string */
	    if (ParsePgm( p1+1, &pgm, &argv )) {
		fprintf( stderr, "Could not parse command line %s\n", 
			p1 + 1);
		break;
		}
	    /* Run program */
	    if (err = PX_Proc_run_program( pgm, argv, 
					  &rd_in, &wrt_out, &wrt_err, 
					  &pgmst )) {
		fprintf( stderr, "Could not run program %s\n", pgm );
		if (err == PX_EXEC_FAILED)
		    exit(1);
		break;
		}
	    PrintPgmArgs( pgm, argv );
	    FreePgmArgs( pgm, argv );
	    /* Attach handlers - wrt_out/err to stdout, rd_in ignored */
	    /* Eventually, we may want to have permanent handlers for this */
	    close_fd	= PX_Handler_simple_create( PX_Hfcn_close_fd, 
						    (void *)0 );
	    to_stdout	= PX_Handler_simple_create( PX_Hfcn_to_stdout, 
						    (void *)0 );
	    err_fd	= PX_Handler_simple_create( PX_Hfcn_error_fd, 
						    (void *)0 );
	    
	    PX_FD_add( wrt_out, 1, to_stdout, err_fd, close_fd, 
		     (struct timeval *)0, ctx );
	    PX_FD_add( wrt_err, 1, to_stdout, err_fd, close_fd, 
		     (struct timeval *)0, ctx );
	    /* It would be nice to report the exit status of this program */
	    }
	    break;
	case 'd':
	    fprintf( stderr, "d (dump) not supported\n" );
	    break;
	default:
	    fprintf( stderr, "Unrecognized command %s\n", cbuf->buf );
	    break;
	}

    /* Move the remainder of buf to the beginning of the array.  Set curlen
       accordingly */
    p1 = cbuf->buf;
    p++;
    while (*p) {
	*p1++ = *p++;
	}
    *p1 = 0;
    cbuf->curlen = (int)(p1 - cbuf->buf);
    }

return 0;
}

/* Read pgm and argv from buf; allocating as required.  argv is null
   terminated */
int ParsePgm( buf, pgm, argv )
char *buf, **pgm, ***argv;
{
char *p1;
int  argc;

while (*buf && isspace(*buf)) buf++;
if (!*buf) return 0;

/* Program name */
p1 = buf;
while (*p1 && !isspace(*p1)) p1++;
*pgm = (char *)MALLOC( (int)(p1 - buf) + 1 );
if (*p1) *p1++ = 0;
if (!*pgm) return -1;
strcpy( *pgm, buf );

/* Make a simple assumption on the number of arguments */
*argv = (char **)MALLOC( 128 * (sizeof(char *)) );
if (!*argv) return -1;

(*argv)[0] = (char *)MALLOC( (int)(p1 - buf) + 1 );
strcpy( (*argv)[0], buf );

/* Program arguments */
argc = 1;
while (1) {
    /* Skip white space */
    buf = p1;
    while (*buf && isspace(*buf)) buf++;
    if (!*buf) break;

    p1 = buf;
    while (*p1 && !isspace(*p1)) p1++;
    (*argv)[argc] = (char *)MALLOC( (int)(p1 - buf) + 1 );
    if (*p1) *p1++ = 0;
    strcpy( (*argv)[argc++], buf );
    }
(*argv)[argc] = 0;

return 0;    
}

void FreePgmArgs( pgm, argv )
char *pgm, **argv;
{
int argc;
FREE( pgm );

argc = 0;
while (argv[argc]) {
    FREE( argv[argc] );
    argc++;
    }
FREE( argv );
}
 
void PrintPgmArgs( pgm, argv )
char *pgm, **argv;
{
int argc;

fprintf( stderr, "-Running %s with argv", pgm );

argc = 0;
while (argv[argc]) {
    fprintf( stderr, " %s,", argv[argc] );
    argc++;
    }
fprintf( stderr, "\n" );
}
