/* 
 * This is a message-passing layer for PEXEC programs.  This is NOT 
 * intended as a general purpose, user-message-passing-system.  Rather,
 * it is a reliable way to send messages to other processes in a 
 * fault-tolerant way.  
 *
 * Note, however, that this interface exploits many of the concepts of 
 * MPI is defining modular, reliable message-passing.
 * However, the "receive" function is handled more in an "active-message"
 * way, as fits the event-driven nature of PEXEC.
 */

#include "pexec.h"
#include <errno.h>
#include "msgpass.h"
/* For htonl/ntohl etc. */
#include <sys/types.h>
#include <netinet/in.h>

#include <stdio.h>
/* 
 * Debugging macros
 */
#define DBG_MSG(mctx,str) {if ((mctx)->dbg_flags) printf( "%s\n",str );}

/*
   MSG_Send - Sends a message to a designated process
 
   Input Parameters:
.  buf - Message in standard (network byte-order) format
.  n   - length of buf in bytes

 */
int MSG_Send( mctx, buf, n, tag, mgroup, rank )
MSG_State *mctx;
void      *buf;
int       n, tag;
MSG_Group mgroup;
int       rank;
{
int        abs_rank, fd;
MSG_Header *header;
MSG_FD     *fdinfo;
MSG_FDout  *fdout;
PX_Ctx     *ctx = mctx->ctx;
PX_Conn    *pxconn;

/* Get the fd to use */
abs_rank = mgroup->l_to_g[rank];
fdinfo   = mctx->fdinfo[abs_rank];

/* If the write is in progress, we have to refuse to add another one */
fdout    = &(fdinfo->fdout);
if (fdout->busy) 
    return MSG_ERR_BUSY;

fd		 = fdout->fd;
/* Note that this fd may be a friend, rather than the ultimate destination */
header		 = &fdout->header;
header->len	 = htonl(n);
header->group_id = htons(mgroup->group_id);
header->msg_type = MSG_DATA;
header->tag	 = tag;
header->source	 = htonl(mgroup->lrank);
header->dest	 = htonl(abs_rank);

/* If write returns without sending all, queue, set writefd bit */
pxconn = &(ctx->conn[fd]);
fdout->n_written = PX_SYSWRITE(pxconn,header,sizeof(MSG_Header)); 
if (fdout->n_written == sizeof(MSG_Header)) {
    fdout->n_sent = PX_SYSWRITE(pxconn,buf,n);
    }
else 
    fdout->n_sent = 0;

if (fdout->n_sent != n ) {
    fdout->buf	   = buf;
    fdout->n_total = n;
    fdout->busy	   = 1;
    /* Flick on writefd mask */
    PX_FD_set_active( ctx, fd, PX_WRITE, 1 );
    }
else {
    fdout->busy    = 0;
    PX_FD_set_active( ctx, fd, PX_WRITE, 0 );
    }

return 0;
}

/* 
   Fd handler routine for message-passing sockets

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.

*/
int MSG_Hfcn_incoming( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
MSG_Header *header;
MSG_FDin   *fdin;
MSG_FD     *fdinfo;
MSG_Group  group;
int        group_id;
int        n;

fdinfo = (MSG_FD *)handler->extra_data;
fdin   = &(fdinfo->fdin);

if (fdin->busy) {
    /* Are we still reading a header? */
    if (fdin->n_read < sizeof(MSG_Header)) {
	n = PX_SYSREAD( conn, ((char *)&fdin->header) + fdin->n_read, 
		       sizeof(MSG_Header) - fdin->n_read );
	if (n > 0) {
	    fdin->n_read += n;
	    }
	else {
	    return PX_Hfcn_do_eof( ctx, conn, handler, n );
	    }
	/* Dispatch the message - we drop through to the code below */
	}
    else {
	/* We must be running the fcn */
	(*fdin->fcn)( fdin->mctx, group, fdin->extra_data );
	/* When to set ? */
	fdin->busy = 0;
	}
    }
else {
    /* Try to read header.  The header is maintained as part of the 
       data on the fd, so we can always read a partial header */
    n = PX_SYSREAD( conn, &fdin->header, sizeof(MSG_Header) );
    if (n > 0) {
	fdin->n_read = n;
	}
    else {
	return PX_Hfcn_do_eof( ctx, conn, handler, n );
	}
    }

/* If have a complete header, find the group and the function array */
if (fdin->n_read < sizeof(MSG_Header)) {
    fdin->busy = 1;
    return 0;
    }

DBG_MSG(fdin->mctx,"Read header; calling function")
/* Call the appropriate function */
group_id = ntohs( fdin->header.group_id );
group    = fdin->mctx->groups[group_id];
if (group->fcn[fdin->header.tag]) {
    fdin->busy       = 1;
    fdin->fcn	     = group->fcn[fdin->header.tag];
    fdin->extra_data = group->extra_data[fdin->header.tag];
    (*fdin->fcn)( fdin->mctx, group, fdin->extra_data );
    }
else {
    /* Unrecognized tag! */
    return 1;
    }
}

/* 
   Fd handler routine for message-passing sockets

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.

  This is used to complete writes of data
*/
int MSG_Hfcn_outgoing( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
MSG_Header *header;
MSG_FDout  *fdout;
MSG_FD     *fdinfo;
fdinfo = (MSG_FD *)handler->extra_data;
fdout  = &(fdinfo->fdout);

if (!fdout->busy) 
    return 0;

if (fdout->n_written < sizeof(MSG_Header)) {
    fdout->n_written += PX_SYSWRITE( conn, 
				  (char*)(&fdout->header) + fdout->n_written, 
				    sizeof(MSG_Header) - fdout->n_written );
    }
if (fdout->n_written == sizeof(MSG_Header)) {
    fdout->n_sent += PX_SYSWRITE( conn, fdout->buf + fdout->n_sent, 
				  fdout->n_total - fdout->n_sent );
    if (fdout->n_sent == fdout->n_total) {
	fdout->busy = 0;
	DBG_MSG(fdout->mctx,"Completed write of message")
	}
    }
return 0;
}



